/*
 * Sketch / Application to write a stream to be a PDP-11 PC11
 * reader interface.
 * 
 *  COPYRIGHT 2018, 2019 Jay R. Jaeger
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 * 
 * PROGRAM LIMITATIONS
 * 
 * Because an Arduino is running software, it cannot always react
 * as quickly as real hardware would.  In particular, BUSY may not
 * be set in time on a normal read operation, and DONE may not
 * get reset quickly enough if multiple read operations are
 * initiated (Reader Enable on the PC11) without an intervening
 * data read operation of the PRB.  This is not expected to affect
 * the bootstrap program or absolute loader, but it might affect
 * paper tape operations in an operating system setting.
 * 
 * Timing is not the same as a real PC11.
 * 
 * PINOUTS
 * 
 * PC11      PDP-11        PC05           Arduino Pin
 * 
 *  D      RDR BUSY L   RDR RUN L           PA0  22
 *  J      BUS INIT L   RDR INITIALIZE L    PA1  23
 *  N      RDR DONE L   IO BUS INT L        PA2  24
 *  R      READ         IOP4(1)             PD1  20 (Interrupt)
 *  T      ERROR        RDR OUTTAPE H       PA4  26
 *  V      SELECT       IOP2(1)             PD0  21 (Interrupt)
 *  Z      BIT 7        IO BUS 04 L         PC7  30
 *  DD     BIT 6        IO BUS 05 L         PC6  31
 *  JJ     BIT 5        IO BUS 06 L         PC5  32
 *  NN     BIT 4        IO BUS 07 L         PC4  33
 *  TT     BIT 3        IO BUS 08 L         PC3  34
 *  RR     BIT 2        IO BUS 09 L         PC2  35
 *  LL     BIT 1        IO BUS 10 L         PC1  36
 *  FF     BIT 0        IO BUS 11 L         PC0  37
 *  F      GND          GND                 N/C
 *  L      1K Pull up   QUALIFY             N/C
 *  X      ----------   IO BUS SKIP L       N/C
 *  BB     BUS INIT H   RDR INITIALIZE H    N/C
 *  HH     ----------   -------------       N/C
 * 
 * Arduino Register Assignments
 * 
 * PORTA  BITS UID.E...   (. == used for other purposes)
 *                        U=RunL, I=InitL, D=DoneL, E=Error
 * PORTB  Bits ........   (. == used for other purposes / unused)
 * PORTC  Bits 76543210
 * PORTD  Bits ......SR   (. == used for other purposes / unused)
 *                        R=Read, S=Select
 * 
 * 
 */

#include <SPI.h>
#include <EthernetV2_0.h>

#define DEBUG 1
#define ONLINE 1
#define ACKMASK  2
#define GETDATA (PINC)

#define W5200_CS 10
#define SDCARD_CS 4
#define TCPPORT 80

#define PIN_RDR_RUN 22
#define PIN_RDR_INIT 23
#define PIN_RDR_DONE 24
#define PIN_READ_IOP4 20
#define PIN_OUT_OF_TAPE 26
#define PIN_SELECT_IOP2 21

#define READER_WAIT_TIME 2     /* Wait this number of ms between chars. */

#define HOSTHEADER "Host: 192.168.42.60"
#define AGENTHEADER "User-Agent: arduino ethernet"
#define CONNECTIONHEADER "Connection: close"


//  Mac Address 0xDE 0xAD 'j' 'r' 'j' #Sequence Number

byte mac[] = { 
  0xDE, 0xAD, 0x6A, 0x72, 0x6A, 0x02 };
  
IPAddress ip(192,168,42,254);
IPAddress server(192,168,42,60);
EthernetClient client;

int debugData[] = {
  0x00, 0xff, 0x01, 0xfe, 0x02, 0xfd, 0x04, 0xfb, 0x08, 0xf7, 
  0x10, 0xef, 0x20, 0xdf, 0x40, 0xbf, 0x80, 0x7f
};

enum states {
  S_START, S_INIT, S_GI, S_READYW4, S_G4, S_READING, S_DATAW2, S_G2, S_ERRORW
};

const char *stateNames[] = {
  "START", "INIT", "GI", "READYW4", "G4", "READING", "DATAW2", "G2", "ERRORW"
};

//  Variables affected or possibly affected by interrupt routines.

volatile bool readerDone = false;            //  True if reader has a character available
volatile bool readerRun = false;             //  True if reader is running tape (busy)
volatile int readerReadIOP4 = 0;             //  State of READ / IOP4 signal from PC11
volatile int readerSelectIOP2 = 0;           //  State of SELECT / IOP2 signal from PC11

int debugging = DEBUG;
int clientConnected = 0;
int dataLength = -1;

void setup() {
  // put your setup code here, to run once:

  //  Set up ports for appropriate directions - mostly IN

  //  PORTA 0, 2, 4 are outputs, 1, 3, 5 are inputs, 6, 7  are unused

  DDRA = DDRA & B11010101;
  DDRA = DDRA | B00010101;
  
  //  PORTC is all outputs

  DDRC = B11111111;

  //  Set up interrupt pins for IOP2 and IOP4 on Port D

  pinMode(PIN_SELECT_IOP2,INPUT);
  pinMode(PIN_READ_IOP4,INPUT);

  attachInterrupt(digitalPinToInterrupt(PIN_SELECT_IOP2),isr_selectIOP2,RISING);
  attachInterrupt(digitalPinToInterrupt(PIN_READ_IOP4),isr_readIOP4,RISING);
  
  //  Set up serial console

  Serial.begin(9600);

  //  Disable the SD Card
  
  pinMode(SDCARD_CS, OUTPUT);
  digitalWrite(SDCARD_CS,HIGH);

  //  If we are online, initialize Ethernet
  
  if(ONLINE) {
    Ethernet.begin(mac,ip);
    delay(1000);
    Serial.print("Arduino PC11 Client initialized at address ");
    Serial.println(Ethernet.localIP());
  }
}

void loop() {

  EthernetClient client;

  bool outOfTape = false;             //  True ff out of tape (error)

  int readerInitL = 0;                //  State of RDR INIT L signal from PC11
  
  bool exiting = false;
  
  int readerWaitCount = READER_WAIT_TIME; /* Time left before DONE *, in ms */
  int debugDataIndex = 0;
  int debugDataSize = sizeof(debugData) / sizeof(debugData[0]);

  enum states state = S_START;
  enum states lastState = S_START;
  
  String lengthHeader = "Content-Length: ";
 
  //  Now, go into a loop.

  while(!exiting) {

    //  Take in the current control signals from the PC11
    //  Saves on a bunch of calls to read pin inputs
    
    readerInitL = digitalRead(PIN_RDR_INIT);
    // readerReadIOP4 = digitalRead(PIN_READ_IOP4);
    // readerSelectIOP2 = digitalRead(PIN_SELECT_IOP2);

    //  Determine the next state, and actions from this state

    //  If we have an INIT signal, all states revert back to S_INIT,
    //  except for S_GI waiting for INIT to go back high.
    //  This code avoids having to test for this in every state.

    if(state != S_GI && readerInitL == 0) {
      state = S_INIT;
    }

    //  Log state changes if debugging.

    if(debugging && state != lastState) {
      Serial.print("State: ");
      Serial.println(stateNames[state]);
    }

    lastState = state;
        
    switch(state) {

      //  S_START:  Waiting for INIT (Active Low)
      
      case S_START:
        if(readerInitL == 0) {
          state = S_INIT;
        }
        break;

      //  S_INIT: Received an INIT L from any of a number
      //  of places.  We cannot send them to S_START because
      //  INIT might go back high before we test again.
      //  Also reset done immeidately as the program could
      //  start right away, and reset the other variables.
            
      case S_INIT:

        PORTA |= B00000100;          //  Clear done immediately (Active LOW)
        readerReadIOP4 = 0;
        readerSelectIOP2 = 0;
        readerRun = false;
        readerDone = false;
        outOfTape = false;
        readerWaitCount = READER_WAIT_TIME;
        debugDataIndex = 0;
        PORTC =  (unsigned byte) ~B10101010;

        delayMicroseconds(5);       //  Let input signal settle
        
        if(debugging) {
            Serial.print(" <-I ");
            Serial.flush(); 
        }
        state = S_GI;        
        break;

      //  S_GI:  Waiting for INIT L to go away
      
      case S_GI:
      
        if(readerInitL == 1) {
          if(debugging) {
            Serial.println(" (gI) ");
            Serial.flush();
          }          
          state = S_READYW4;
        }
        break;

      //  S_READYW4:  Waiting for READ / IOP4 indicating
      //  the start of a read.  IOP2 would be OK - just reading data again
        
      case S_READYW4:

        // if(readerSelectIOP2 == 1) {
        //  readerSelectIOP2 = 0;
        //  state = S_G2;
        //  break;
        // }

        //  Is the PC11 asking for us to read a character?  If so, throw
        //  away any character we might already have in waiting, and start
        //  the read process.

        //  Note that the interrupt service routine turns off done and turns
        //  on busy immediately.
      
        if(readerReadIOP4 == 1) {
          delayMicroseconds(5);     //  Debounce / let input signal settle                        
          if(debugging) {
            Serial.print(" <-R ");   
            Serial.flush();       
          }
          
          //  Set the appropriate status bits for the PC11
          //  (This also happens in the interrupt routine.  Doing
          //  this here is actually superfluous, but not harmful)
          
          readerDone = false;
          readerRun = true;
          
          //  Do we have anything more to send?

          //  If we are debugging and not online, use the data array 
          //  index to decide.  Otherwise see if we have some data in 
          //  TCP to send to the PC11.

          if(!ONLINE && debugging && debugDataIndex >= debugDataSize) {            
            outOfTape = true;
            readerRun = false;
          }
          if(ONLINE && !clientConnected) {
            outOfTape = true;          
            readerRun = false;          
          }

          if(debugging && outOfTape) {
            Serial.print(" O-> ");
            Serial.flush();
          }      

          readerReadIOP4 = 0;
          state = S_G4;
        }    
            
        break;

        //  State S_G4:  Waiting for the IOP4 to go away.
        //  INIT resets, IOP2 is an error (should not be asserted at same time)

        //  If we are out of tape, go back to waiting - so that we don't set
        //  DONE.

        //  Once IOP4 goes away, if we are out of tape, state goes
        //  back to waiting for IOP4 (and PC11 will see ERROR bit).
        //  Otherwise we go into the reading delay state.

      case S_G4:

        if(readerReadIOP4 == 0) {          
          delayMicroseconds(5);             //  Debounce delay
          if(outOfTape) {
            state = S_READYW4;
          }
          else {
            readerWaitCount = READER_WAIT_TIME;
            readerRun = true;
            readerDone = false;
            state = S_READING;
          }
        }

        if(readerSelectIOP2 == 1) {
          Serial.println();
          Serial.println("Unexpected IOP2 waiting for NOT IOP4 in state S_G4");
          Serial.flush();
          readerSelectIOP2 = 0;
          state = S_ERRORW;
          break;
        }        
        
        break;

      //  State S_READING.  Delay before indicating data is ready.
      //  An IOP4 during this time sends us back to S_READYW4.
      //  Data is left unchanged (it would be indeterminate on a real reader)
      //  An IOP2 during this time just returns old data, so ignore it.
      //  If the wait time expires, we make the data available, and
      //  move on to wait for the IOP2.
      
      case S_READING:

        // if(readerReadIOP4 == 1) {
        //   readerDone = false;
        //   readerRun = false;          
        //   readerReadIOP4 = 0;
        //   state = S_READYW4;
        //   break;
        // }

        if(readerWaitCount <= 0) {
          
          //  WAIT to clear BUSY and set DONE until
          //  we have set the data port.  Otherwise,
          //  the PDP11 may react to quickly, and get
          //  stale data.
          
          if(debugging && !ONLINE) {
            //  Return fake data if debugging.
            if(debugDataIndex < debugDataSize) {
              PORTC = ~debugData[debugDataIndex];
              ++debugDataIndex;              
              Serial.print(" d-> ");
              Serial.print(debugData[debugDataIndex],HEX);
              Serial.print(" (W2) ");
              Serial.flush();              
            }
          }
          else {
            if(ONLINE && clientConnected) {
              if(dataLength > 0) {
                int c = client.read();
                if(debugging) {
                  Serial.print(" D-> ");
                  Serial.print(c,HEX);
                  Serial.print(" (W2) ");
                  Serial.flush();
                }
                PORTC = ~c;
                delayMicroseconds(5);     //  Settling delay
                --dataLength;
                outOfTape = false;
              }
              else {
                client.stop();
                clientConnected = 0;
                outOfTape = true;    
                if(debugging) {
                  Serial.print(" OO-> ");
                  Serial.flush();
                }
              }
            }
          }

          readerRun = false;
          readerDone= true;       //  This should be the ONLY place done is SET
          state = S_DATAW2;          
        }
        else {
          --readerWaitCount;
        }
        
        break;

      //  State S_DATAW2: Waiting for an IOP2.  An IOP4 during this time
      //  causes us to ignore the character, and go back to state S_READYW4
      //  An IOP2 causes us to send the data.

      //  In this state it is possible for the PDP-11 to get ahead of us, and
      //  issue the READ (IOP4) before we are done with IOP2.  Just leave the
      //  variable set, and continue on.
        
      case S_DATAW2:

        // if(readerReadIOP4 == 1) {
        //  readerDone = false;
        //  state = S_READYW4;        //  Leave readerReadIOP4 set at 1 in this case.
        //  break;
        // }

        if(readerSelectIOP2 == 1) {  
          delayMicroseconds(5);       //  Debounce        
          if(debugging) {
            Serial.print(" <-S ");
            Serial.flush();
          }
          
          readerDone = false;     
          readerSelectIOP2 = 0;                        
          state = S_G2;
        }
        
        break;

      //  State S_G2: Wait for IOP2 to be withdrawn.  
        
      case S_G2:

        if(readerSelectIOP2 == 0) {
          state = S_READYW4;
          delayMicroseconds(5);             //  Debounce
          if(debugging) {
            Serial.println(" (g2) ");
            Serial.flush();
          }          
        }
        
        break;

      //  Error state.  Wait for an INIT (handled above), or both
      //  IOP2 and IOP4 to go low, and reset as though there was an init.
      
      case S_ERRORW:
        if(readerInitL == 0 || 
           (readerReadIOP4 == 0 && readerSelectIOP2 == 0)) {
          state = S_GI;
        }
        break;
 
      default:
        Serial.print("Invalid State: ");
        Serial.println(state);
        Serial.flush();
        state = S_ERRORW;
        break;
    }

    //  Send status to the PC11

    digitalWrite(PIN_RDR_RUN,readerRun ? 0 : 1);      //  Active Low
    digitalWrite(PIN_RDR_DONE,readerDone ? 0 : 1);    //  Active Low
    digitalWrite(PIN_OUT_OF_TAPE,outOfTape ? 1 : 0);  //  Active High

    //  See if we have any instructions from our master...

    while(Serial.available() > 0) {
      int incoming = Serial.read();
      switch(incoming) {
        case 'd':
        case 'D':
          debugging = true;
          Serial.println("Enter debugging mode.");
          Serial.flush();
          break;
          
        case 'l':
        case 'L':
          
          //  If we are operating online, then get the file
          //  and connect.

          if(ONLINE) {           

            String fileName;
            String request;
            int inHeaders = true;

            fileName = Serial.readStringUntil('\n');
            request = "GET /" + fileName + " HTTP/1.1";            
            
            if(clientConnected) {
              if(debugging) {
                Serial.println("Previous connection closing.");
              }
              client.stop();
            }

            if(debugging) {
              Serial.println("Connecting to server to send " + request);
              Serial.flush();
            }
            
            if(client.connect(server,TCPPORT)) {
              if(debugging){ 
                Serial.println("Connected.");
                Serial.flush();            
              }
              clientConnected = 1;
              client.println(request);
              client.println(HOSTHEADER);
              client.println(AGENTHEADER);
              client.println(CONNECTIONHEADER);
              client.println();
              outOfTape = false;

              //  Process response headers, looking for content-length
              //  which tells us how much data we can read.  Also, clear
              //  out of tape / Error if there is data.

              dataLength = -1;
              while(inHeaders) {
                if(client.available()) {
                  String header = client.readStringUntil('\n');
                  if(debugging) {
                    Serial.print("Header: " + header + ", Length: ");
                    Serial.println(header.length());                                              
                  }                  
                  if(header.length() < 2) {
                    inHeaders = false;
                    break;
                  }
                  if(header.startsWith(lengthHeader)) {
                    dataLength = header.substring(lengthHeader.length()).toInt();                    
                    if(debugging) {
                      Serial.print(lengthHeader + " header found.  Length is: ");
                      Serial.println(dataLength);                        
                    }
                    if(dataLength > 0) {
                      outOfTape = false;
                    }
                  }
                }
              }              
            }
            else {
              Serial.println("Unable to connect to server.  Offline.");
              clientConnected = 0;
              outOfTape = true;       
            }
          }
          else if(debugging) {                    
            outOfTape = false;
          }

          readerReadIOP4 = 0;
          readerSelectIOP2 = 0;
          
          break;
          
        case 'n':
        case 'N':
          debugging = false;
          Serial.println("Exit debugging mode");
          Serial.flush();
          break;
        case 'x':
        case 'X':
          exiting = true;
          break;
        case 'v':
        case 'V':
          Serial.print("OOT: "); Serial.print(outOfTape);
          Serial.print(", Done: "); Serial.print(readerDone);
          Serial.print(", Run: "); Serial.print(readerRun);
          Serial.print(", Wait: "); Serial.print(readerWaitCount);
          Serial.println();
          Serial.print("INIT L: "); Serial.print(readerInitL);
          Serial.print(", IOP2: "); Serial.print(readerSelectIOP2);
          Serial.print(", IOP4: "); Serial.print(readerReadIOP4);
          Serial.println();
          Serial.print("Digital pins: IOP2: "); 
          Serial.print(digitalRead(PIN_SELECT_IOP2));
          Serial.print(", IOP4: ");
          Serial.println(digitalRead(PIN_READ_IOP4));
          if(debugging) {
            Serial.print(" Data Index: "); Serial.println(debugDataIndex);            
          }
          Serial.flush();
        default:
          ;
      }

    }

    //  If the reader is running, wait for 1ms

    if(state == S_READING) {
      delay(1);
    }
      
  }     //  End Main loop
    
  if(ONLINE && clientConnected) {
    client.stop();
  }

  Serial.println();
  Serial.println("Done. ");
}

//  Interrupt Service Routines (ISRs)

void isr_selectIOP2() {

    //  Clear DONE immediately (Active LOW)

    PORTA |= B00000100;
    readerDone = false;
    readerRun = false;

    //  Tell the state machine about the IOP2
    
    readerSelectIOP2 = 1;

}

void isr_readIOP4() {

  //  Most PDP-11 programs test for DONE, so
  //  clearing that is the top priority - needs
  //  to be done directly.  (Active LOW)

  PORTA |= B00000100;
  
  //  Set Run/Busy directly is also priority, but
  //  few programs seem to actually use it. 
  //  (Active LOW)

  PORTA &= B11111110;

  //  Set the flags for the rest of the program.
  
  readerDone = false;
  readerRun = true;

  //  Tell the state machine about the IOP4 flag.
  readerReadIOP4 = 1; 

  //  Also clear IOP2 flag.  It cannot have been
  //  set while we were in this ISR. ;)

  //  readerSelectIOP2 = 0;  
}
